#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <fcntl.h>
#ifdef TURBOC
#include <alloc.h>
#endif
#include "wirefram.h"

METHODS Bezier_Methods = {
   Bezier_Displayer, Bezier_Evaluater, Bezier_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Ellipse_Methods = {
   Generic_Displayer, Ellipse_Evaluater, Ellipse_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Hyperboloid_Methods = {
   Generic_Displayer, Hyperboloid_Evaluater, Hyperboloid_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Hyperbolic_Paraboloid_Methods = {
   Generic_Displayer, Hyperbolic_Paraboloid_Evaluater,
   Hyperbolic_Paraboloid_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Paraboloid_Methods = {
   Generic_Displayer, Paraboloid_Evaluater, Paraboloid_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Cone_Methods = {
   Generic_Displayer, Cone_Evaluater, Cone_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Plane_Methods = {
   Generic_Displayer, Plane_Evaluater, Plane_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Coil_Methods = {
   Generic_Displayer, Coil_Evaluater, Coil_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Squishy_Methods = {
   Generic_Displayer, Squishy_Evaluater, Squishy_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Torus_Methods = {
   Generic_Displayer, Torus_Evaluater, Torus_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Polygon_Methods = {
   Polygon_Displayer, Polygon_Evaluater, Polygon_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Radial_Methods = {
   Radial_Displayer, Radial_Evaluater, Radial_Normaler,
   Generic_Rotater1, Generic_Rotater2, Generic_Translater,
   Generic_Scaler};

METHODS Viewpoint_Methods = {
   NULL, NULL, NULL,
   Viewpoint_Rotater1, Viewpoint_Rotater1,
   Generic_Translater, Generic_Scaler};

ORIENTATION Default_Orientation = {
    {0.0, 0.0, 0.0}, {0.0, 1.0, 0.0}, {0.0, 0.0, 1.0},
    {1.0, 0.0, 0.0}, {1.0, 1.0, 1.0}};
ORIENTATION Default_Light_Orientation = {
    {100.0, 100.0, -100.0}, {0.0, 0.0, 1.0}, {0.0, 1.0, 0.0},
    {1.0, 0.0, 0.0}, {1.0, 1.0, 1.0}};
ORIENTATION Default_Viewpoint_Orientation = {
   {0.0, 0.0, -10.0}, {0.0, 0.0, 1.0}, {0.0, 1.0, 0.0},
    {1.0, 0.0, 0.0}, {1.0, 1.0, 1.0}};

PARAMETERS Default_Ellipse_Parameters = { 0.0, 1.0, 0.0, 1.0, 8, 8 };
PARAMETERS Default_Hyperboloid_Parameters = { 0.0, 1.0, 0.0, 1.0, 8, 8 };
PARAMETERS Default_HyperParab_Parameters = { 0.3, 0.7, 0.3, 0.7, 8, 8 };
PARAMETERS Default_Paraboloid_Parameters = { 0.0, 1.0, 0.0, 1.0, 8, 8 };
PARAMETERS Default_Cone_Parameters = { 0.0, 1.0, 0.0, 1.0, 8, 8 };
PARAMETERS Default_Plane_Parameters = {-1.0, 1.0,-1.0, 1.0, 4, 4 };
PARAMETERS Default_Torus_Parameters = { 0.0, 1.0, 0.0, 1.0, 8, 16 };
PARAMETERS Default_Coil_Parameters = { 0.0, 1.0, 0.0, 1.0, 16, 8 };
PARAMETERS Default_Squishy_Parameters = { 0.0, 1.0, 0.0, 1.0, 16, 16 };
PARAMETERS Default_Bezier_Parameters = { 0.0, 1.0, 0.0, 1.0, 3, 3 };
PARAMETERS Default_Radial_Parameters = { 0.0, 1.0, 0.0, 1.0, 16, 36 };

VECTOR tensor1[4][4] = {
   {{-1.5,-1.0,-1.5}, {-0.5,-1.5,-1.5}, { 0.5, 1.5,-1.5}, { 1.5, 1.0,-1.5}},
   {{-1.5, 0.5,-0.5}, {-0.5, 0.2,-0.5}, { 0.5,-0.2,-0.5}, { 1.5,-0.5,-0.5}},
   {{-1.5, 0.5, 0.5}, {-0.5, 0.2, 0.5}, { 0.5,-0.2, 0.5}, { 1.5,-0.5, 0.5}},
   {{-1.5,-1.0, 1.5}, {-0.5,-1.5, 1.5}, { 0.5, 1.5, 1.5}, { 1.5, 1.0, 1.5}}};

BEZIER_PARAMS Default_Bezier_Params = { &tensor1, 0.1 };
FOCUS Default_Viewpoint_Focus = {{0.0, 0.0, 0.0}, 10.0, 0.46365, 1.0};

/* Create a new instance of an object, filling in default parameters */
OBJECT *
create_object(object_type)
   int object_type;
{
   int i, j;
   OBJECT *New_Object = NULL;

   New_Object = (OBJECT *)malloc(sizeof(OBJECT));
   if (New_Object == NULL) {
      printf("Failed to allocate object\n");
      exit(1);
      return NULL;
      }
   New_Object->Next_Object = NULL;
   if (object_type != Polygon_Type &&
       object_type != Viewpoint_Type &&
       object_type != Light_Source_Type) {
      New_Object->Parameters = (PARAMETERS *)malloc(sizeof(PARAMETERS));
      if (New_Object->Parameters == NULL) {
         printf("Failed to allocate parameters\n");
         exit(1);
         free(New_Object);
         return NULL;
         }
      }
   else
      New_Object->Parameters = NULL;
   New_Object->Object_Type = object_type;
   New_Object->Orientation = Default_Orientation;
   New_Object->Dummy = NULL;
   New_Object->Texture.Color[0] = 1.0;
   New_Object->Texture.Color[1] = 1.0;
   New_Object->Texture.Color[2] = 1.0;
   New_Object->Texture.Color[3] = 0.0;
   New_Object->Texture.Phong = 1.0;

   switch (object_type) {
      case Viewpoint_Type:
         New_Object->Orientation = Default_Viewpoint_Orientation;
         New_Object->Methods = &Viewpoint_Methods;
         ((VIEWPOINT *)New_Object)->Focus = &Default_Viewpoint_Focus;
         break;
      case Light_Source_Type:
         New_Object->Orientation = Default_Light_Orientation;
         New_Object->Methods = NULL;
         break;
      case Bezier_Type:
         New_Object->Methods = &Bezier_Methods;
         *New_Object->Parameters = Default_Bezier_Parameters;
         ((BEZIER *)New_Object)->Bezier_Params = &Default_Bezier_Params;
         break;
      case Ellipse_Type:
         New_Object->Methods = &Ellipse_Methods;
         *New_Object->Parameters = Default_Ellipse_Parameters;
         break;
      case Hyperboloid_Type:
         New_Object->Methods = &Hyperboloid_Methods;
         *New_Object->Parameters = Default_Hyperboloid_Parameters;
         break;
      case Cone_Type:
         New_Object->Methods = &Cone_Methods;
         *New_Object->Parameters = Default_Cone_Parameters;
         break;
      case Paraboloid_Type:
         New_Object->Methods = &Paraboloid_Methods;
         *New_Object->Parameters = Default_Paraboloid_Parameters;
         break;
      case Plane_Type:
         New_Object->Methods = &Plane_Methods;
         *New_Object->Parameters = Default_Plane_Parameters;
         break;
      case Torus_Type:
         New_Object->Methods = &Torus_Methods;
         *New_Object->Parameters = Default_Torus_Parameters;
         ((TORUS *)New_Object)->Torus_Params =
            (TORUS_PARAMS *)malloc(sizeof(TORUS_PARAMS));
         ((TORUS *)New_Object)->Torus_Params->a = 0.3;
         ((TORUS *)New_Object)->Torus_Params->k = 0.7;
         break;
      case Coil_Type:
         New_Object->Methods = &Coil_Methods;
         *New_Object->Parameters = Default_Coil_Parameters;
         ((COIL *)New_Object)->Coil_Params =
            (COIL_PARAMS *)malloc(sizeof(COIL_PARAMS));
         ((COIL *)New_Object)->Coil_Params->a = 0.3;
         ((COIL *)New_Object)->Coil_Params->c = 1.0;
         ((COIL *)New_Object)->Coil_Params->k = 1.0;
         break;
      case Hyperbolic_Paraboloid_Type:
         New_Object->Methods = &Hyperbolic_Paraboloid_Methods;
         *New_Object->Parameters = Default_HyperParab_Parameters;
         New_Object->Object_Type = Hyperbolic_Paraboloid_Type;
         break;
      case Radial_Type:
         New_Object->Methods = &Radial_Methods;
         *New_Object->Parameters = Default_Radial_Parameters;
         ((RADIAL *)New_Object)->Radial_Params =
            (RADIAL_PARAMS *)malloc(sizeof(RADIAL_PARAMS));
         ((RADIAL *)New_Object)->Radial_Params->Sector_Length =
            Default_Sector_Length;
         ((RADIAL *)New_Object)->Radial_Params->Sector_Length =
            Default_Radial_Display;
         for (i=0;i<Default_Radial_Count;i++)
            for (j=0;j<Default_Sector_Count;j++)
               ((RADIAL *)New_Object)->Radial_Params->plat[i][j] = 0;
         break;
      case Polygon_Type:
         New_Object->Methods = &Polygon_Methods;
         New_Object->Parameters = NULL;
         ((POLYGON *)New_Object)->Polygon_Params = NULL;
         break;
      case Squishy_Type:
         New_Object->Methods = &Squishy_Methods;
         *New_Object->Parameters = Default_Squishy_Parameters;
         ((SQUISHY *)New_Object)->Squishy_Params =
            (SQUISHY_PARAMS *)malloc(sizeof(SQUISHY_PARAMS));
         ((SQUISHY *)New_Object)->Squishy_Params->a = 2.0;
         ((SQUISHY *)New_Object)->Squishy_Params->b = 2.0;
         ((SQUISHY *)New_Object)->Squishy_Params->c = 1.0;
         break;
      default:
         printf("Bad object type: %d\n", object_type);
         exit(1);
         }
   return New_Object;
}

void
dump_triangle(dump_file, v0, v1, v2, n0, n1, n2)
   FILE *dump_file;
   VECTOR *v0, *v1, *v2;
   VECTOR *n0, *n1, *n2;
{
   DBL l0;
   VECTOR vx;

   /* Eliminate degenerate triangles */
   VSub(vx, *v0, *v1); VLength(l0, vx);
   if (fabs(l0) < EPSILON) return;
   VSub(vx, *v0, *v2); VLength(l0, vx);
   if (fabs(l0) < EPSILON) return;
   VSub(vx, *v1, *v2); VLength(l0, vx);
   if (fabs(l0) < EPSILON) return;

   if (n0 == NULL || n1 == NULL || n2 == NULL) {
      fprintf(dump_file, "object\n");
      fprintf(dump_file, "   triangle");
      fprintf(dump_file, " <%.3f %.3f %.3f>", v0->x, v0->y, v0->z);
      fprintf(dump_file, " <%.3f %.3f %.3f>", v1->x, v1->y, v1->z);
      fprintf(dump_file, " <%.3f %.3f %.3f>", v2->x, v2->y, v2->z);
      fprintf(dump_file, " end_triangle\n");
      fprintf(dump_file, "      texture This_Texture end_texture\n");
      fprintf(dump_file, "end_object\n\n");
      }
   else {
      fprintf(dump_file, "object\n");
      fprintf(dump_file, "   smooth_triangle");
      fprintf(dump_file, " <%.3f %.3f %.3f> <%.3f %.3f %.3f>",
              v0->x, v0->y, v0->z, n0->x, n0->y, n0->z);
      fprintf(dump_file, " <%.3f %.3f %.3f> <%.3f %.3f %.3f>",
              v1->x, v1->y, v1->z, n1->x, n1->y, n1->z);
      fprintf(dump_file, " <%.3f %.3f %.3f> <%.3f %.3f %.3f>",
              v2->x, v2->y, v2->z, n2->x, n2->y, n2->z);
      fprintf(dump_file, " end_smooth_triangle\n");
      fprintf(dump_file, "   texture This_Texture end_texture\n");
      fprintf(dump_file, "end_object\n\n");
      }
}

void
check_min_max(minv, maxv, testv)
   VECTOR *minv, *maxv, *testv;
{
   if (minv->x > testv->x) minv->x = testv->x;
   if (minv->y > testv->y) minv->y = testv->y;
   if (minv->z > testv->z) minv->z = testv->z;
   if (maxv->x < testv->x) maxv->x = testv->x;
   if (maxv->y < testv->y) maxv->y = testv->y;
   if (maxv->z < testv->z) maxv->z = testv->z;
}

void
DKB_Dump(Object, dump_file, smooth_flag)
   OBJECT *Object;
   FILE *dump_file;
   int smooth_flag;
{
   VECTOR minvec, maxvec;
   DBL maxz, sx, sy, sz, sr;
   int i, j, k;
   DBL u, v, x, y, z, param[2], *param_ptr;
   VECTOR v0, v1;
   DBL u_min = Object->Parameters->Min_U;
   DBL u_max = Object->Parameters->Max_U;
   DBL v_min = Object->Parameters->Min_V;
   DBL v_max = Object->Parameters->Max_V;
   int u_steps = Object->Parameters->U_Steps;
   int v_steps = Object->Parameters->V_Steps;
   DBL u_range, v_range, delta_u, delta_v;
   VECTOR vertices[4], normals[4];

   param_ptr = &param[0];

   u_range = u_max - u_min;
   v_range = v_max - v_min;

   /* Calculate the grid and project it. */
   delta_u = u_range / (DBL)u_steps;
   delta_v = v_range / (DBL)v_steps;

   /* Dump out triangles */
   for (i=0,u=u_min;i<u_steps;i++,u+=delta_u) {
      fprintf(dump_file, "composite\n");
      minvec.x = 1.0e10; maxvec.x = -1.0e10;
      minvec.y = 1.0e10; maxvec.y = -1.0e10;
      minvec.z = 1.0e10; maxvec.z = -1.0e10;
      for (j=0,v=v_min;j<v_steps;j++,v+=delta_v) {
         param[0] = u; param[1] = v;
         Evaluate(Object, param_ptr, &v0);
         Normal(Object, param_ptr, &normals[0]);
         world_position_object(Object, &v0, &vertices[0]);
         param[1] = v+delta_v;
         Evaluate(Object, param_ptr, &v0);
         Normal(Object, param_ptr, &normals[1]);
         world_position_object(Object, &v0, &vertices[1]);
         param[0] = u+delta_u;
         Evaluate(Object, param_ptr, &v0);
         Normal(Object, param_ptr, &normals[2]);
         world_position_object(Object, &v0, &vertices[2]);
         param[1] = v;
         Evaluate(Object, param_ptr, &v0);
         Normal(Object, param_ptr, &normals[3]);
         world_position_object(Object, &v0, &vertices[3]);

         check_min_max(&minvec, &maxvec, &vertices[0]);
         check_min_max(&minvec, &maxvec, &vertices[1]);
         check_min_max(&minvec, &maxvec, &vertices[2]);
         check_min_max(&minvec, &maxvec, &vertices[3]);

         if (smooth_flag) {
            dump_triangle(dump_file,
                          &vertices[0], &vertices[1], &vertices[3],
                          &normals[0], &normals[1], &normals[3]);
            dump_triangle(dump_file,
                          &vertices[1], &vertices[2], &vertices[3],
                          &normals[1], &normals[2], &normals[3]);
            }
         else {
            dump_triangle(dump_file,
                          &vertices[0], &vertices[1], &vertices[3],
                          NULL, NULL, NULL);
            dump_triangle(dump_file,
                          &vertices[1], &vertices[2], &vertices[3],
                          NULL, NULL, NULL);
            }
         }
      sx = (minvec.x + maxvec.x) / 2.0;
      sy = (minvec.y + maxvec.y) / 2.0;
      sz = (minvec.z + maxvec.z) / 2.0;
      sr = sqrt((maxvec.x - sx) * (maxvec.x - sx) +
                (maxvec.y - sy) * (maxvec.y - sy) +
                (maxvec.z - sz) * (maxvec.z - sz));
      fprintf(dump_file, "   bounded_by\n");
      fprintf(dump_file, "      sphere <%.3f %.3f %.3f> %.3f end_sphere\n",
              sx, sy, sz, sr);
      fprintf(dump_file, "   end_bound\n");
      fprintf(dump_file, "end_composite\n");
      }
}

void
rtriangle(dump_file, v0, v1, v2, n0, n1, n2)
   FILE *dump_file;
   VECTOR *v0, *v1, *v2;
   VECTOR *n0, *n1, *n2;
{
   DBL l0;
   VECTOR vx;

   /* Eliminate degenerate triangles */
   VSub(vx, *v0, *v1); VLength(l0, vx);
   if (fabs(l0) < EPSILON) return;
   VSub(vx, *v0, *v2); VLength(l0, vx);
   if (fabs(l0) < EPSILON) return;
   VSub(vx, *v1, *v2); VLength(l0, vx);
   if (fabs(l0) < EPSILON) return;

   if (n0 == NULL || n1 == NULL || n2 == NULL) {
      fprintf(dump_file, "t");
      fprintf(dump_file, " %.3f %.3f %.3f", v0->x, v0->y, v0->z);
      fprintf(dump_file, " %.3f %.3f %.3f", v1->x, v1->y, v1->z);
      fprintf(dump_file, " %.3f %.3f %.3f\n", v2->x, v2->y, v2->z);
      }
   else {
      fprintf(dump_file, "n");
      fprintf(dump_file, " %.3f %.3f %.3f %.3f %.3f %.3f",
              v0->x, v0->y, v0->z, n0->x, n0->y, n0->z);
      fprintf(dump_file, " %.3f %.3f %.3f %.3f %.3f %.3f",
              v1->x, v1->y, v1->z, n1->x, n1->y, n1->z);
      fprintf(dump_file, " %.3f %.3f %.3f %.3f %.3f %.3f\n",
              v2->x, v2->y, v2->z, n2->x, n2->y, n2->z);
      }
}

void
triangle_dump(Object, dump_file, smooth_flag)
   OBJECT *Object;
   FILE *dump_file;
   int smooth_flag;
{
   VECTOR minvec, maxvec;
   DBL maxz, sx, sy, sz, sr;
   int i, j, k;
   DBL u, v, x, y, z, param[2], *param_ptr;
   VECTOR v0, v1;
   DBL u_min = Object->Parameters->Min_U;
   DBL u_max = Object->Parameters->Max_U;
   DBL v_min = Object->Parameters->Min_V;
   DBL v_max = Object->Parameters->Max_V;
   int u_steps = Object->Parameters->U_Steps;
   int v_steps = Object->Parameters->V_Steps;
   DBL u_range, v_range, delta_u, delta_v;
   VECTOR vertices[4], normals[4], new_normals[4];
   VECTOR Angles, Axes[3];

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   param_ptr = &param[0];

   u_range = u_max - u_min;
   v_range = v_max - v_min;

   /* Calculate the grid and project it. */
   delta_u = u_range / (DBL)u_steps;
   delta_v = v_range / (DBL)v_steps;

   /* Dump out triangles */
   for (i=0,u=u_min;i<u_steps;i++,u+=delta_u) {
      for (j=0,v=v_min;j<v_steps;j++,v+=delta_v) {
         param[0] = u; param[1] = v;
         Evaluate(Object, param_ptr, &v0);
         Normal(Object, param_ptr, &normals[0]);
         rotate_points(&Angles, 1, &new_normals[0], &normals[0]);
         world_position_object(Object, &v0, &vertices[0]);
         param[1] = v+delta_v;
         Evaluate(Object, param_ptr, &v0);
         Normal(Object, param_ptr, &normals[1]);
         rotate_points(&Angles, 1, &new_normals[1], &normals[1]);
         world_position_object(Object, &v0, &vertices[1]);
         param[0] = u+delta_u;
         Evaluate(Object, param_ptr, &v0);
         Normal(Object, param_ptr, &normals[2]);
         rotate_points(&Angles, 1, &new_normals[2], &normals[2]);
         world_position_object(Object, &v0, &vertices[2]);
         param[1] = v;
         Evaluate(Object, param_ptr, &v0);
         Normal(Object, param_ptr, &normals[3]);
         rotate_points(&Angles, 1, &new_normals[3], &normals[3]);
         world_position_object(Object, &v0, &vertices[3]);

         check_min_max(&minvec, &maxvec, &vertices[0]);
         check_min_max(&minvec, &maxvec, &vertices[1]);
         check_min_max(&minvec, &maxvec, &vertices[2]);
         check_min_max(&minvec, &maxvec, &vertices[3]);

         if (smooth_flag) {
            rtriangle(dump_file,
                          &vertices[0], &vertices[1], &vertices[3],
                          &normals[0], &normals[1], &normals[3]);
            rtriangle(dump_file,
                          &vertices[1], &vertices[2], &vertices[3],
                          &normals[1], &normals[2], &normals[3]);
            }
         else {
            rtriangle(dump_file,
                          &vertices[0], &vertices[1], &vertices[3],
                          NULL, NULL, NULL);
            rtriangle(dump_file,
                          &vertices[1], &vertices[2], &vertices[3],
                          NULL, NULL, NULL);
            }
         }
      }
}

void
get_orientation(Object)
   OBJECT *Object;
{
   char text_line[256];
   int fields;
   float newx, newy, newz;
   VECTOR Angles, Axes[3], New_Axes[3];
   ORIENTATION *Orientation = &Object->Orientation;

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   /* Display & get new values */
   printf("\nObject Origin: {%f, %f, %f}: ",
          Orientation->Origin.x,
          Orientation->Origin.y,
          Orientation->Origin.z);
   if (gets(&text_line[0]) != NULL) {
      fields = sscanf(text_line, "%f %f %f", &newx, &newy, &newz);
      if (fields == 3) {
         Orientation->Origin.x = newx;
         Orientation->Origin.y = newy;
         Orientation->Origin.z = newz;
         }
      }
   printf("Object Scale: {%f, %f, %f} : ",
          Orientation->Scale.x,
          Orientation->Scale.y,
          Orientation->Scale.z);
   if (gets(&text_line[0]) != NULL) {
      fields = sscanf(text_line, "%f %f %f", &newx, &newy, &newz);
      if (fields == 3) {
         Orientation->Scale.x = newx;
         Orientation->Scale.y = newy;
         Orientation->Scale.z = newz;
         }
      }
   printf("Object Rotations: {%f, %f, %f}: ",
          radians_to_degrees(Angles.x),
          radians_to_degrees(Angles.y),
          radians_to_degrees(Angles.z));
   if (gets(&text_line[0]) != NULL) {
      fields = sscanf(text_line, "%f %f %f", &newx, &newy, &newz);
      if (fields == 3) {
         /* Modify the orientation of the object */
         Angles.x = degrees_to_radians(newx);
         Angles.y = degrees_to_radians(newy);
         Angles.z = degrees_to_radians(newz);
         Axes[0].x = 0; Axes[0].y = 0; Axes[0].z = 1;
         Axes[1].x = 0; Axes[1].y = 1; Axes[1].z = 0;
         Axes[2].x = 1; Axes[2].y = 0; Axes[2].z = 0;
         rotate_points(&Angles, 3, &New_Axes[0], &Axes[0]);
         Object->Orientation.Major_Axis  = New_Axes[0];
         Object->Orientation.Minor_Axis1 = New_Axes[1];
         Object->Orientation.Minor_Axis2 = New_Axes[2];
         }
      }
}

void
edit_light(Object)
   OBJECT *Object;
{
   char text_line[256];
   int fields;
   float newx, newy, newz;

   /* Display & get new values */
   printf("\nLight Origin: {%f, %f, %f}: ",
          Object->Orientation.Origin.x,
          Object->Orientation.Origin.y,
          Object->Orientation.Origin.z);
   if (gets(&text_line[0]) != NULL) {
      fields = sscanf(text_line, "%f %f %f", &newx, &newy, &newz);
      if (fields == 3) {
         Object->Orientation.Origin.x = newx;
         Object->Orientation.Origin.y = newy;
         Object->Orientation.Origin.z = newz;
         }
      }
}

/* The space for the parameters must already exist. */
void
get_parameters(Object)
   OBJECT *Object;
{
   char temps[256];
   float tempval;
   int tempint;
   PARAMETERS *Parameters = Object->Parameters;

   if (Parameters == NULL) return;

   if (Object->Object_Type == Coil_Type ||
       Object->Object_Type == Squishy_Type) {
      printf("Minimum U value: %f        : ", Object->Parameters->Min_U);
      if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
	 Object->Parameters->Min_U = tempval;
      printf("Maximum U value: %f        : ", Object->Parameters->Max_U);
      if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
	 Object->Parameters->Max_U = tempval;
      printf("Minimum V value: %f        : ", Object->Parameters->Min_V);
      if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
	 Object->Parameters->Min_V = tempval;
      printf("Maximum V value: %f        : ", Object->Parameters->Max_V);
      if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
	 Object->Parameters->Max_V = tempval;
      }
   printf("Steps in U     : %d        : ", Object->Parameters->U_Steps);
   if (gets(temps) != NULL) if (sscanf(temps, "%d", &tempint) == 1)
      if (tempint >0 && tempint < MAX_U_STEPS)
         Object->Parameters->U_Steps = tempint;
   printf("Steps in V     : %d        : ", Object->Parameters->V_Steps);
   if (gets(temps) != NULL) if (sscanf(temps, "%d", &tempint) == 1)
      if (tempint >0 && tempint < MAX_V_STEPS)
         Object->Parameters->V_Steps = tempint;
}

/* The space for the parameters must already exist. */
void
get_colors(Object)
   OBJECT *Object;
{
   DBL newcolor[4];
   DBL phongval;
   char temps[256];
   int fields;

   printf("Colors: RED %f green %f blue %f alpha %f: ",
          Object->Texture.Color[0],
          Object->Texture.Color[1],
          Object->Texture.Color[2],
          Object->Texture.Color[3]);
   if (gets(temps) != NULL) {
      fields = sscanf(temps, "%f %f %f %f\n",
                      &newcolor[0], &newcolor[1],
                      &newcolor[2], &newcolor[3]);
      if (fields >= 3 && fields <= 4) {
         Object->Texture.Color[0] = newcolor[0];
         Object->Texture.Color[1] = newcolor[1];
         Object->Texture.Color[2] = newcolor[2];
         if (fields == 4)
            Object->Texture.Color[3] = newcolor[3];
         }
      }
   printf("Phong: %f: ", Object->Texture.Phong);
   if (gets(temps) != NULL) {
      fields = sscanf(temps, "%d", &phongval);
      if (fields == 1 && phongval >= 0.0 && phongval <= 1.0)
         Object->Texture.Phong = phongval;
      }
}

/* The space for the unique parameters must already exist. */
void
Torus_Unique_Entry(Object)
   OBJECT *Object;
{
   DBL tempval;
   char temps[80];
   TORUS_PARAMS *Params = (TORUS_PARAMS *)Object->Dummy;

   printf("Value of a: %f : ", Params->a);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->a = tempval;
   printf("Value of k: %f : ", Params->k);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->k = tempval;
}

void
Coil_Unique_Entry(Object)
   OBJECT *Object;
{
   DBL tempval;
   char temps[80];
   COIL_PARAMS *Params = (COIL_PARAMS *)Object->Dummy;

   printf("Value of a: %f : ", Params->a);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->a = tempval;
   printf("Value of c: %f : ", Params->c);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->c = tempval;
   printf("Value of k: %f : ", Params->k);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->k = tempval;
}

void
Squishy_Unique_Entry(Object)
   OBJECT *Object;
{
   DBL tempval;
   char temps[80];
   SQUISHY_PARAMS *Params = (SQUISHY_PARAMS *)Object->Dummy;

   printf("Value of a: %f : ", Params->a);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->a = tempval;
   printf("Value of b: %f : ", Params->b);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->b = tempval;
   printf("Value of c: %f : ", Params->c);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->c = tempval;
}

void
Viewpoint_Unique_Entry(Object)
   OBJECT *Object;
{
   DBL tempval, lr, lu, x, y, z;
   char temps[80];
   FOCUS Params;
   VECTOR v0;
   VIEWPOINT *Viewpoint = (VIEWPOINT *)Object;
   FOCUS *Focus = (FOCUS *)Object->Dummy;

   printf("Field of view    : %f : ", Focus->Field_Of_View);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1) {
      Focus->Field_Of_View = tempval;
      }
   printf("Aspect Ratio: %f : ", Focus->Aspect);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1) {
      Focus->Aspect = tempval;
      VLength(lu, Viewpoint->Orientation.Minor_Axis1);
      VLength(lr, Viewpoint->Orientation.Minor_Axis2);
      VScale(Viewpoint->Orientation.Minor_Axis2,
             Viewpoint->Orientation.Minor_Axis2, tempval * lu / lr);
      }
   printf("Look at: <%f, %f, %f> : ",
          Focus->Look_At.x, Focus->Look_At.y, Focus->Look_At.z);
   if (gets(temps) != NULL)
      if (sscanf(temps, "%f %f %f", &x, &y, &z) == 3) {
         Focus->Look_At.x = x;
         Focus->Look_At.y = y;
         Focus->Look_At.z = z;
         }
}

void
Polygon_Unique_Entry(poly)
   POLYGON *poly;
{
   char text_line[80];
   DBL newx, newy, newz;
   DBL tempval;
   VECTOR *vertices;
   int i, fields, vertex_count;
   char temps[80], ch;

   /* Display & get new values */
   l0:
   printf("How many segments? ");
   if (gets(text_line) != NULL) {
      fields = sscanf(text_line, "%d", &vertex_count);
      if (fields == 1)
         if (vertex_count >= 1 && vertex_count < MAX_SIDES_PER_POLYGON) {
            poly->Polygon_Params->Vertex_Count = vertex_count;
            vertices = (VECTOR *)malloc(vertex_count * sizeof(VECTOR));
            if (vertices == NULL) {
               printf("Failed to allocate the segment list\n");
               }
            }
         else
            goto l0;
      else
         goto l0;
      }

   for (i=0;i<vertex_count;i++) {
      l1:
      printf("Vertex (%2d): ");
      if (gets(text_line) != NULL) {
         fields = sscanf(text_line, "%f %f %f", &newx, &newy, &newz);
         if (fields == 3) {
            vertices[i].x = newx;
            vertices[i].y = newy;
            vertices[i].z = newz;
            }
         else {
            printf("Must be 3 coordinates per vertex\n");
            goto l1;
            }
         }
      else {
         printf("Null entry, repeat\n");
         goto l1;
         }
      }
   poly->Polygon_Params->Vertices = vertices;
   printf("Is the polygon closed [Y]/N?");
   ch = getch();
   if (ch == 'n' || ch == 'N')
      poly->Polygon_Params->Close_Flag = 0;
   else
      poly->Polygon_Params->Close_Flag = 1;
}

void
Radial_Unique_Entry(Object)
   OBJECT *Object;
{
   int i, tempint;
   DBL tempval;
   char temps[80];
   RADIAL_PARAMS *Params;
   RADIAL *Radial = (RADIAL *)Object;

   Params = Radial->Radial_Params;

   l0:
   printf("Sector Length: %g : ", Params->Sector_Length);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1) {
      if (tempval > 0 && tempval < 10)
         Radial->Radial_Params->Sector_Length = tempval;
      else
         goto l0;
      }
   l1:
   printf("Display Type: %d : ", Params->Display_Type);
   if (gets(temps) != NULL) if (sscanf(temps, "%d", &tempint) == 1) {
      if (tempint >= 0 && tempint < 2)
         Radial->Radial_Params->Display_Type = tempint;
      else
         goto l1;
      }
   l2:
   printf("Radial # (cr to end): ");
   if (gets(temps) != NULL) if (sscanf(temps, "%d", &tempint) == 1) {
      if (tempint >= 0 && tempint < Default_Radial_Count) {
         printf("Enter sector: ");
         if (gets(temps) != NULL)
            if (strlen(temps) == Default_Sector_Count) {
               for (i=0;i<Default_Sector_Count;i++)
                  if (temps[i] == '0')
                     Radial->Radial_Params->plat[tempint][i] = 0;
                  else if (temps[i] == '1')
                     Radial->Radial_Params->plat[tempint][i] = 1;
                  else
                     printf("Bad value at sector %d: '%c'\n", i, temps[i]);
               goto l2;
               }
            else {
               printf("Wrong # of sectors: %d, must be %d\n",
                      strlen(temps), Default_Sector_Count);
               goto l2;
               }
         else
            goto l2;
         }
      }
}

void
Bezier_Unique_Entry(Object)
   OBJECT *Object;
{
   DBL tempval;
   char temps[80];
   BEZIER_PARAMS *Params = (BEZIER_PARAMS *)Object->Dummy;

   printf("Subdivision Epsilon: %f : ", Params->Division_Epsilon);
   if (gets(temps) != NULL) if (sscanf(temps, "%f", &tempval) == 1)
      Params->Division_Epsilon = tempval;
}

void
get_unique_params(Object)
   OBJECT *Object;
{
   switch (Object->Object_Type) {
   case Bezier_Type:
      Bezier_Unique_Entry(Object);
      break;
   case Torus_Type:
      Torus_Unique_Entry(Object);
      break;
   case Coil_Type:
      Coil_Unique_Entry(Object);
      break;
   case Squishy_Type:
      Squishy_Unique_Entry(Object);
      break;
   case Polygon_Type:
      Polygon_Unique_Entry((POLYGON *)Object);
      break;
   case Radial_Type:
      Radial_Unique_Entry(Object);
      break;
   case Viewpoint_Type:
      Viewpoint_Unique_Entry(Object);
      break;
   }
}

void
Plane_DKB_Dumper(Object, dkb_file)
   OBJECT *Object;
   FILE *dkb_file;
{
   VECTOR Angles, Axes[3];

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   fprintf(dkb_file, "{ Define a plane. }\n");
   fprintf(dkb_file, "object\n");
   fprintf(dkb_file, "   plane <%f %f %f> 0 end_plane\n",
           Axes[0].x, Axes[0].y, Axes[0].z);
   fprintf(dkb_file, "   scale <%f %f %f>\n",
           Object->Orientation.Scale.x,
           Object->Orientation.Scale.y,
           Object->Orientation.Scale.z);
   if (fabs(Object->Orientation.Origin.x) > 0.0001 ||
       fabs(Object->Orientation.Origin.y) > 0.0001 ||
       fabs(Object->Orientation.Origin.z) > 0.0001)
      fprintf(dkb_file, "   translate <%f %f %f>\n",
              Object->Orientation.Origin.x,
              Object->Orientation.Origin.y,
              Object->Orientation.Origin.z);
   fprintf(dkb_file, "   texture\n");
   fprintf(dkb_file, "      color red %.3f green %.3f blue %.3f",
           Object->Texture.Color[0],
           Object->Texture.Color[1],
           Object->Texture.Color[2]);
   if (Object->Texture.Color[3] > 0.0)
      fprintf(dkb_file, " alpha %d\n",
              Object->Texture.Color[3]);
   else
      fprintf(dkb_file, "\n");
   fprintf(dkb_file, "      phong %f\n",
           Object->Texture.Phong);
   fprintf(dkb_file, "   END_texture\n");
   fprintf(dkb_file, "end_object\n\n");
}

/* Dump a torus with the specified size, location, orientation, and texture. */
void
Torus_DKB_Dumper(Object, dkb_file)
   OBJECT *Object;
   FILE *dkb_file;
{
   DBL Major_Radius, Minor_Radius;
   DBL br, r0, r1, l0, l1, l2;
   VECTOR Angles, Axes[3];
   TORUS_PARAMS *tp = ((TORUS *)Object)->Torus_Params;

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   Major_Radius = tp->k;
   Minor_Radius = tp->a;
   r0 = Major_Radius * Major_Radius;
   r1 = Minor_Radius * Minor_Radius;
   l0 = -2.0 * (r0 + r1);
   l1 = 2.0 * (r0 - r1);
   l2 = (r0 - r1) * (r0 - r1);
   br = (Major_Radius + Minor_Radius + 0.1);
   br = br * br;

   fprintf(dkb_file,
           "{ Define a torus with major radius %.2f, minor radius %.2f. }\n",
           Major_Radius, Minor_Radius);
   fprintf(dkb_file, "object\n");
   fprintf(dkb_file, "   quartic\n");
   fprintf(dkb_file, "      < 1.0 0.0 0.0 0.0 2.0\n");
   fprintf(dkb_file, "        0.0 0.0 2.0 0.0 %.4f\n", l0);
   fprintf(dkb_file, "        0.0 0.0 0.0 0.0 0.0\n");
   fprintf(dkb_file, "        0.0 0.0 0.0 0.0 0.0\n");
   fprintf(dkb_file, "        1.0 0.0 0.0 2.0 0.0\n");
   fprintf(dkb_file, "        %.4f 0.0 0.0 0.0 0.0\n", l1);
   fprintf(dkb_file, "        1.0 0.0 %.4f 0.0 %.4f>\n", l0, l2);
   fprintf(dkb_file, "      rotate <0 0 -90>\n");
   fprintf(dkb_file, "   end_quartic\n");
   fprintf(dkb_file, "   bounded_by quadric \n");
   fprintf(dkb_file, "           <1.0 1.0 1.0>\n");
   fprintf(dkb_file, "           <0.0 0.0 0.0>\n");
   fprintf(dkb_file, "           <0.0 0.0 0.0>\n");
   fprintf(dkb_file, "           -%f\n", br);
   fprintf(dkb_file, "      end_quadric\n");
   fprintf(dkb_file, "   end_bound\n");
   if (fabs(Angles.z) > 0.0001)
      fprintf(dkb_file, "   rotate <0 0 %f>\n",
              radians_to_degrees(Angles.z));
   if (fabs(Angles.x) > 0.0001 || fabs(Angles.y) > 0.0001)
      fprintf(dkb_file, "   rotate <%f %f 0>\n",
              radians_to_degrees(Angles.x),
              radians_to_degrees(Angles.y));
   if (fabs(Object->Orientation.Origin.x) > 0.0001 ||
       fabs(Object->Orientation.Origin.y) > 0.0001 ||
       fabs(Object->Orientation.Origin.z) > 0.0001)
      fprintf(dkb_file, "   translate <%f %f %f>\n",
              Object->Orientation.Origin.x,
              Object->Orientation.Origin.y,
              Object->Orientation.Origin.z);
   fprintf(dkb_file, "   texture\n");
   fprintf(dkb_file, "      color red %.3f green %.3f blue %.3f",
           Object->Texture.Color[0],
           Object->Texture.Color[1],
           Object->Texture.Color[2]);
   if (Object->Texture.Color[3] > 0.0)
      fprintf(dkb_file, " alpha %d\n",
              Object->Texture.Color[3]);
   else
      fprintf(dkb_file, "\n");
   fprintf(dkb_file, "      phong %f\n",
           Object->Texture.Phong);
   fprintf(dkb_file, "   END_texture\n");
   fprintf(dkb_file, "end_object\n\n");
}

void
Ellipse_DKB_Dumper(Object, dkb_file)
   OBJECT *Object;
   FILE *dkb_file;
{
   VECTOR Angles, Axes[3];

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   fprintf(dkb_file, "{ Define an Ellipsoid. }\n");
   fprintf(dkb_file, "object\n");
   fprintf(dkb_file, "   quadric <1.0 1.0 1.0>\n");
   fprintf(dkb_file, "           <0.0 0.0 0.0>\n");
   fprintf(dkb_file, "           <0.0 0.0 0.0>\n");
   fprintf(dkb_file, "           -1.0\n");
   fprintf(dkb_file, "   end_quadric\n");
   fprintf(dkb_file, "   scale <%f %f %f>\n",
           Object->Orientation.Scale.x,
           Object->Orientation.Scale.y,
           Object->Orientation.Scale.z);
   if (fabs(Angles.z) > 0.0001)
      fprintf(dkb_file, "   rotate <0 0 %f>\n",
              radians_to_degrees(Angles.z));
   if (fabs(Angles.x) > 0.0001 || fabs(Angles.y) > 0.0001)
      fprintf(dkb_file, "   rotate <%f %f 0>\n",
              radians_to_degrees(Angles.x),
              radians_to_degrees(Angles.y));
   if (fabs(Object->Orientation.Origin.x) > 0.0001 ||
       fabs(Object->Orientation.Origin.y) > 0.0001 ||
       fabs(Object->Orientation.Origin.z) > 0.0001)
      fprintf(dkb_file, "   translate <%f %f %f>\n",
              Object->Orientation.Origin.x,
              Object->Orientation.Origin.y,
              Object->Orientation.Origin.z);
   fprintf(dkb_file, "   texture\n");
   fprintf(dkb_file, "      color red %.3f green %.3f blue %.3f",
           Object->Texture.Color[0],
           Object->Texture.Color[1],
           Object->Texture.Color[2]);
   if (Object->Texture.Color[3] > 0.0)
      fprintf(dkb_file, " alpha %d\n",
              Object->Texture.Color[3]);
   else
      fprintf(dkb_file, "\n");
   fprintf(dkb_file, "      phong %f\n",
           Object->Texture.Phong);
   fprintf(dkb_file, "   END_texture\n");
   fprintf(dkb_file, "end_object\n\n");
}

void
Hyperboloid_DKB_Dumper(Object, dkb_file)
   OBJECT *Object;
   FILE *dkb_file;
{
   VECTOR Angles, Axes[3];

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   fprintf(dkb_file, "{ Define a hyperboloid of 1 sheet. }\n");
   fprintf(dkb_file, "object\n");
   fprintf(dkb_file, "   intersection\n");
   fprintf(dkb_file, "      quadric <1.0  1.0 -1.0>\n");
   fprintf(dkb_file, "              <0.0  0.0  0.0>\n");
   fprintf(dkb_file, "              <0.0  0.0  0.0>\n");
   fprintf(dkb_file, "              -1.0\n");
   fprintf(dkb_file, "      end_quadric\n");
   fprintf(dkb_file, "      plane <0 0  1> %.3f end_plane\n",
           fabs(sinh(1.76274717 * (0.5 - Object->Parameters->Max_U))));
   fprintf(dkb_file, "      plane <0 0 -1> %.3f end_plane\n",
           fabs(sinh(1.76274717 * (0.5 - Object->Parameters->Min_U))));
   fprintf(dkb_file, "   end_intersection\n");
   fprintf(dkb_file, "   scale <%.3f %.3f %.3f>\n",
           Object->Orientation.Scale.x,
           Object->Orientation.Scale.y,
           Object->Orientation.Scale.z);
   if (fabs(Angles.z) > EPSILON)
      fprintf(dkb_file, "   rotate <0 0 %.3f>\n",
              radians_to_degrees(Angles.z));
   if (fabs(Angles.x) > EPSILON || fabs(Angles.y) > EPSILON)
      fprintf(dkb_file, "   rotate <%.3f %.3f 0>\n",
              radians_to_degrees(Angles.x),
              radians_to_degrees(Angles.y));
   fprintf(dkb_file, "   translate <%.3f %.3f %.3f>\n",
           Object->Orientation.Origin.x,
           Object->Orientation.Origin.y,
           Object->Orientation.Origin.z);
   fprintf(dkb_file, "   texture\n");
   fprintf(dkb_file, "      color red %.3f green %.3f blue %.3f",
           Object->Texture.Color[0],
           Object->Texture.Color[1],
           Object->Texture.Color[2]);
   if (Object->Texture.Color[3] > 0.0)
      fprintf(dkb_file, " alpha %.3f\n",
              Object->Texture.Color[3]);
   else
      fprintf(dkb_file, "\n");
   fprintf(dkb_file, "      phong %f\n",
           Object->Texture.Phong);
   fprintf(dkb_file, "   END_texture\n");
   fprintf(dkb_file, "end_object\n\n");
}

void
Paraboloid_DKB_Dumper(Object, dkb_file)
   OBJECT *Object;
   FILE *dkb_file;
{
   VECTOR Angles, Axes[3];

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   fprintf(dkb_file, "{ Define a paraboloid. }\n");
   fprintf(dkb_file, "object\n");
   fprintf(dkb_file, "   intersection\n");
   fprintf(dkb_file, "      quadric <1.0  1.0  0.0>\n");
   fprintf(dkb_file, "              <0.0  0.0  0.0>\n");
   fprintf(dkb_file, "              <0.0  0.0 -1.0>\n");
   fprintf(dkb_file, "              0.0\n");
   fprintf(dkb_file, "      end_quadric\n");
   fprintf(dkb_file, "      plane <0 0  1> %f end_plane\n",
           Object->Parameters->Max_V * Object->Parameters->Max_V);
   fprintf(dkb_file, "   end_intersection\n");
   fprintf(dkb_file, "   scale <%.3f %.3f %.3f>\n",
           Object->Orientation.Scale.x,
           Object->Orientation.Scale.y,
           Object->Orientation.Scale.z);
   fprintf(dkb_file, "   rotate <0 0 %.3f>\n",
           radians_to_degrees(Angles.z));
   fprintf(dkb_file, "   rotate <%.3f %.3f 0>\n",
           radians_to_degrees(Angles.x),
           radians_to_degrees(Angles.y));
   fprintf(dkb_file, "   translate <%.3f %.3f %.3f>\n",
           Object->Orientation.Origin.x,
           Object->Orientation.Origin.y,
           Object->Orientation.Origin.z);
   fprintf(dkb_file, "   texture\n");
   fprintf(dkb_file, "      color red %.3f green %.3f blue %.3f",
           Object->Texture.Color[0],
           Object->Texture.Color[1],
           Object->Texture.Color[2]);
   if (Object->Texture.Color[3] > 0.0)
      fprintf(dkb_file, " alpha %.3f\n",
              Object->Texture.Color[3]);
   else
      fprintf(dkb_file, "\n");
   fprintf(dkb_file, "      phong %.3f\n",
           Object->Texture.Phong);
   fprintf(dkb_file, "   END_texture\n");
   fprintf(dkb_file, "end_object\n\n");
}

void
Cone_DKB_Dumper(Object, dkb_file)
   OBJECT *Object;
   FILE *dkb_file;
{
   VECTOR Angles, Axes[3];

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   fprintf(dkb_file, "{ Define a (half) cone. }\n");
   fprintf(dkb_file, "object\n");
   fprintf(dkb_file, "   intersection\n");
   fprintf(dkb_file, "      quadric <1.0  1.0 -1.0>\n");
   fprintf(dkb_file, "              <0.0  0.0  0.0>\n");
   fprintf(dkb_file, "              <0.0  0.0  0.0>\n");
   fprintf(dkb_file, "              0.0\n");
   fprintf(dkb_file, "      end_quadric\n");
   fprintf(dkb_file, "      plane <0 0 -1> 0 end_plane\n");
   fprintf(dkb_file, "      plane <0 0  1> %f end_plane\n",
           Object->Parameters->Max_V * Object->Parameters->Max_V);
   fprintf(dkb_file, "   end_intersection\n");
   fprintf(dkb_file, "   scale <%.3f %.3f %.3f>\n",
           Object->Orientation.Scale.x,
           Object->Orientation.Scale.y,
           Object->Orientation.Scale.z);
   fprintf(dkb_file, "   rotate <0 0 %.3f>\n",
           radians_to_degrees(Angles.z));
   fprintf(dkb_file, "   rotate <%.3f %.3f 0>\n",
           radians_to_degrees(Angles.x),
           radians_to_degrees(Angles.y));
   fprintf(dkb_file, "   translate <%.3f %.3f %.3f>\n",
           Object->Orientation.Origin.x,
           Object->Orientation.Origin.y,
           Object->Orientation.Origin.z);
   fprintf(dkb_file, "   texture\n");
   fprintf(dkb_file, "      color red %.3f green %.3f blue %.3f",
           Object->Texture.Color[0],
           Object->Texture.Color[1],
           Object->Texture.Color[2]);
   if (Object->Texture.Color[3] > 0.0)
      fprintf(dkb_file, " alpha %.3f\n",
              Object->Texture.Color[3]);
   else
      fprintf(dkb_file, "\n");
   fprintf(dkb_file, "      phong %.3f\n",
           Object->Texture.Phong);
   fprintf(dkb_file, "   end_texture\n");
   fprintf(dkb_file, "end_object\n\n");
}

void
Bezier_DKB_Dumper(Object, dkb_file)
   OBJECT *Object;
   FILE *dkb_file;
{
   VECTOR Angles, Axes[3];

   Axes[0] = Object->Orientation.Major_Axis;
   Axes[1] = Object->Orientation.Minor_Axis1;
   Axes[2] = Object->Orientation.Minor_Axis2;
   determine_reference_angles(&Axes[0], &Angles);

   fprintf(dkb_file, "{ Define a Bezier patch. }\n");
   fprintf(dkb_file, "object\n");
   fprintf(dkb_file, "   scale <%.3f %.3f %.3f>\n",
           Object->Orientation.Scale.x,
           Object->Orientation.Scale.y,
           Object->Orientation.Scale.z);
   fprintf(dkb_file, "   rotate <0 0 %.3f>\n",
           radians_to_degrees(Angles.z));
   fprintf(dkb_file, "   rotate <%.3f %.3f 0>\n",
           radians_to_degrees(Angles.x),
           radians_to_degrees(Angles.y));
   fprintf(dkb_file, "   translate <%.3f %.3f %.3f>\n",
           Object->Orientation.Origin.x,
           Object->Orientation.Origin.y,
           Object->Orientation.Origin.z);
   fprintf(dkb_file, "   texture\n");
   fprintf(dkb_file, "      color red %.3f green %.3f blue %.3f",
           Object->Texture.Color[0],
           Object->Texture.Color[1],
           Object->Texture.Color[2]);
   if (Object->Texture.Color[3] > 0.0)
      fprintf(dkb_file, " alpha %.3f\n",
              Object->Texture.Color[3]);
   else
      fprintf(dkb_file, "\n");
   fprintf(dkb_file, "      phong %.3f\n",
           Object->Texture.Phong);
   fprintf(dkb_file, "   end_texture\n");
   fprintf(dkb_file, "end_object\n\n");
}

void
Viewpoint_DKB_Dumper(Eye, dkb_file)
   VIEWPOINT *Eye;
   FILE *dkb_file;
{
   VECTOR right;
   VScale(right, Eye->Orientation.Minor_Axis2, Eye->Focus->Aspect);

   fprintf(dkb_file, "view_point\n");
   fprintf(dkb_file, "   location <%f %f %f>\n",
           Eye->Orientation.Origin.x,
           Eye->Orientation.Origin.y,
           Eye->Orientation.Origin.z);
   fprintf(dkb_file, "   direction <%f %f %f>\n",
           Eye->Orientation.Major_Axis.x,
           Eye->Orientation.Major_Axis.y,
           Eye->Orientation.Major_Axis.z);
   fprintf(dkb_file, "   up <%f %f %f>\n",
           Eye->Orientation.Minor_Axis1.x,
           Eye->Orientation.Minor_Axis1.y,
           Eye->Orientation.Minor_Axis1.z);
   fprintf(dkb_file, "   right <%f %f %f>\n", right.x, right.y, right.z);
   fprintf(dkb_file, "end_view_point\n\n");
}

void
Light_DKB_Dumper(OBJECT *Light_Source, FILE *dkb_file)
{
   fprintf(dkb_file, "object\n");
   fprintf(dkb_file, "   sphere <0 0 0> 1 end_sphere\n");
   fprintf(dkb_file, "   translate <%f %f %f>\n",
           Light_Source->Orientation.Origin.x,
           Light_Source->Orientation.Origin.y,
           Light_Source->Orientation.Origin.z);
   fprintf(dkb_file, "   texture\n");
   fprintf(dkb_file, "      color White\n");
   fprintf(dkb_file, "      ambient 1.0\n");
   fprintf(dkb_file, "      diffuse 0.0\n");
   fprintf(dkb_file, "   end_texture\n");
   fprintf(dkb_file, "   light_source\n");
   fprintf(dkb_file, "   color White\n");
   fprintf(dkb_file, "end_object\n\n");
}

void
dump_DKB_object(Object, dkb_file)
   OBJECT *Object;
   FILE *dkb_file;
{
   /* Dump any information specific to this type of object */
   switch (Object->Object_Type) {
   case Bezier_Type:
      Bezier_DKB_Dumper(Object, dkb_file);
      break;
   case Plane_Type:
      Plane_DKB_Dumper(Object, dkb_file);
      break;
   case Torus_Type:
      Torus_DKB_Dumper(Object, dkb_file);
      break;
   case Ellipse_Type:
      Ellipse_DKB_Dumper(Object, dkb_file);
      break;
   case Cone_Type:
      Cone_DKB_Dumper(Object, dkb_file);
      break;
   case Paraboloid_Type:
      Paraboloid_DKB_Dumper(Object, dkb_file);
      break;
   case Hyperboloid_Type:
      Hyperboloid_DKB_Dumper(Object, dkb_file);
      break;
   default:
      DKB_Dump(Object, dkb_file, 1);
   }
}

void
write_DKB_information(dkb_file, Eye, Light_Sources, All_Objects)
   FILE *dkb_file;
   VIEWPOINT *Eye;
   OBJECT *Light_Sources;
   OBJECT *All_Objects;
{
   OBJECT *Object, *Lights;

   fprintf(dkb_file, "INCLUDE \"colors.dat\"\n");
   fprintf(dkb_file, "INCLUDE \"textures.dat\"\n");
   fprintf(dkb_file, "INCLUDE \"shapes.dat\"\n");

   for (Object = All_Objects;
        Object != NULL;
        Object = Object->Next_Object)
      dump_DKB_object(Object, dkb_file);

   Viewpoint_DKB_Dumper(Eye, dkb_file);

   for (Lights = Light_Sources;
        Lights != NULL;
        Lights = Lights->Next_Object)
      Light_DKB_Dumper(Lights, dkb_file);
}

void
write_triangle_information(file, All_Objects, flag)
   FILE *file;
   OBJECT *All_Objects;
   int flag;
{
   OBJECT *Object;
   for (Object = All_Objects;
        Object != NULL;
        Object = Object->Next_Object)
      triangle_dump(Object, file, flag);
}


void
Torus_Wire_Dumper(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   fprintf(wire_file, "%g %g\n",
           ((TORUS *)Object)->Torus_Params->a,
           ((TORUS *)Object)->Torus_Params->k);
}

void
Torus_Wire_Reader(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   TORUS *Torus = (TORUS *)Object;
   char text_line[80];
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g",
          &Torus->Torus_Params->a,
          &Torus->Torus_Params->k);
}

void
Coil_Wire_Dumper(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   fprintf(wire_file, "%g %g %g\n",
           ((COIL *)Object)->Coil_Params->a,
           ((COIL *)Object)->Coil_Params->c,
           ((COIL *)Object)->Coil_Params->k);
}

void
Coil_Wire_Reader(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   COIL *Coil = (COIL *)Object;
   char text_line[80];
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g %g",
          &Coil->Coil_Params->a,
          &Coil->Coil_Params->c,
          &Coil->Coil_Params->k);
}

void
Squishy_Wire_Dumper(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   fprintf(wire_file, "%g %g %g\n",
           ((SQUISHY *)Object)->Squishy_Params->a,
           ((SQUISHY *)Object)->Squishy_Params->b,
           ((SQUISHY *)Object)->Squishy_Params->c);
}

void
Squishy_Wire_Reader(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   SQUISHY *Squishy = (SQUISHY *)Object;
   char text_line[80];
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g %g",
          &Squishy->Squishy_Params->a,
          &Squishy->Squishy_Params->b,
          &Squishy->Squishy_Params->c);
}

void
Polygon_Wire_Dumper(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   int i, cnt;
   VECTOR *Vertices;

   Vertices = ((POLYGON *)Object)->Polygon_Params->Vertices;
   cnt = ((POLYGON *)Object)->Polygon_Params->Vertex_Count;
   fprintf(wire_file, "%d %d\n",
           cnt, ((POLYGON *)Object)->Polygon_Params->Close_Flag);
   for (i=0;i<cnt;i++)
      fprintf(wire_file, "%g %g %g\n",
              Vertices[i].x, Vertices[i].y, Vertices[i].z);
}

void
Polygon_Wire_Reader(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   int i, cnt;
   VECTOR *Vertices;
   char text_line[80];

   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%d %d",
          &cnt, &(((POLYGON *)Object)->Polygon_Params->Close_Flag));
   ((POLYGON *)Object)->Polygon_Params->Vertex_Count = cnt;
   Vertices = (VECTOR *)malloc(cnt * sizeof(VECTOR));
   if (Vertices == NULL) {
      printf("Failed to allocate polygon vertices\n");
      ((POLYGON *)Object)->Polygon_Params->Vertices = NULL;
      return;
      }
   for (i=0;i<cnt;i++) {
      fgets(&text_line[0], 255, wire_file);
      sscanf(text_line, "%g %g %g",
             &Vertices[i].x, &Vertices[i].y, &Vertices[i].z);
      }
   ((POLYGON *)Object)->Polygon_Params->Vertices = Vertices;
}

void
Radial_Wire_Reader(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   int i, j;
   char text_line[255];

   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%d",
          &((RADIAL *)Object)->Radial_Params->Display_Type);
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g",
          &((RADIAL *)Object)->Radial_Params->Sector_Length);

   for (i=0;i<Default_Radial_Count;i++) {
      fgets(&text_line[0], 255, wire_file);
      for (j=0;j<Default_Sector_Count;j++)
         ((RADIAL *)Object)->Radial_Params->plat[i][j] = text_line[j] - '0';
      }
}

void
Radial_Wire_Dumper(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   int i, j;
   fprintf(wire_file, "%d\n",
           ((RADIAL *)Object)->Radial_Params->Display_Type);
   fprintf(wire_file, "%g\n",
           ((RADIAL *)Object)->Radial_Params->Sector_Length);
   for (i=0;i<Default_Radial_Count;i++) {
      for (j=0;j<Default_Sector_Count;j++)
         fprintf(wire_file, "%c",
            ((RADIAL *)Object)->Radial_Params->plat[i][j] + '0');
      fprintf(wire_file, "\n");
      }
}

void
Bezier_Wire_Reader(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   int i, j;
   char text_line[255];
   VECTOR v0, (*Patch)[4][4];
   BEZIER_PARAMS *Params;

   Params = (BEZIER_PARAMS *)malloc(sizeof(BEZIER_PARAMS));

   /* By default, a predefined mesh is used. Allocate space for the
      custom one being read. */
   Patch = (void *)malloc((4 * 4) * sizeof(VECTOR));
   if (Patch == NULL) {
      printf("Couldn't allocate space for the patch\n");
      exit(1);
      }
   Params->Patch = Patch;

   /* Get the flatness criteria */
   fgets(&text_line[0], 255, wire_file);
   sscanf(&text_line[0], "%f", &(Params->Division_Epsilon));

   /* Read in the individual patch coordinates, 1 per line. */
   for (i=0;i<4;i++) {
      for (j=0;j<4;j++) {
         fgets(&text_line[0], 255, wire_file);
         sscanf(&text_line[0], "%f %f %f", &v0.x, &v0.y, &v0.z);
         (*Patch)[i][j] = v0;
         }
      }
   ((BEZIER *)Object)->Bezier_Params = Params;
}

void
Bezier_Wire_Dumper(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   int i, j;
   BEZIER_PARAMS *Params;
   VECTOR (*Patch)[4][4];

   Params = ((BEZIER *)Object)->Bezier_Params;
   Patch = Params->Patch;

   /* Write the flatness criteria */
   fprintf(wire_file, "%f\n", Params->Division_Epsilon);

   /* Write in the individual patch coordinates, 1 per line. Row major order. */
   for (i=0;i<4;i++) {
      for (j=0;j<4;j++) {
         fprintf(wire_file, "%f %f %f\n",
                 (*Patch)[i][j].x, (*Patch)[i][j].y, (*Patch)[i][j].z);
         }
      }
}

void
Viewpoint_Wire_Dumper(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   int i, j;
   VIEWPOINT *View = (VIEWPOINT *)Object;
   FOCUS *Focus = View->Focus;

   /* Write the look at and distance values */
   fprintf(wire_file, "%g %g %g %g %g %g\n",
           Focus->Look_At.x, Focus->Look_At.y, Focus->Look_At.z,
           Focus->Distance, Focus->Field_Of_View, Focus->Aspect);
}

void
Viewpoint_Wire_Reader(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   char text_line[255];
   VIEWPOINT *View = (VIEWPOINT *)Object;
   FOCUS *Focus;

   Focus = (FOCUS *)malloc(sizeof(FOCUS));
   if (Focus == NULL) {
      printf("Failed to allocate viewpoint\n");
      exit(1);
      }
   else
      View->Focus = Focus;

   fgets(&text_line[0], 255, wire_file);
   sscanf(&text_line[0], "%g %g %g %g %g %g",
          &(Focus->Look_At.x), &(Focus->Look_At.y), &(Focus->Look_At.z),
          &(Focus->Distance), &(Focus->Field_Of_View), &(Focus->Aspect));
   printf("New focus: (%g %g %g), dist: %g\n",
          Focus->Look_At.x, Focus->Look_At.y, Focus->Look_At.z,
          Focus->Distance);
}

void
wire_read_information(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   char text_line[256];

   /* Read data common to all object types */
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g %g", 
          &Object->Orientation.Origin.x,
          &Object->Orientation.Origin.y,
          &Object->Orientation.Origin.z);
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g %g", 
          &Object->Orientation.Major_Axis.x,
          &Object->Orientation.Major_Axis.y,
          &Object->Orientation.Major_Axis.z);
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g %g", 
          &Object->Orientation.Minor_Axis1.x,
          &Object->Orientation.Minor_Axis1.y,
          &Object->Orientation.Minor_Axis1.z);
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g %g", 
          &Object->Orientation.Minor_Axis2.x,
          &Object->Orientation.Minor_Axis2.y,
          &Object->Orientation.Minor_Axis2.z);
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g %g", 
          &Object->Orientation.Scale.x,
          &Object->Orientation.Scale.y,
          &Object->Orientation.Scale.z);
   fgets(&text_line[0], 255, wire_file);
   sscanf(text_line, "%g %g %g %g %g\n",
           &Object->Texture.Color[0],
           &Object->Texture.Color[1],
           &Object->Texture.Color[2],
           &Object->Texture.Color[3],
           &Object->Texture.Phong);
   fgets(&text_line[0], 255, wire_file);
   if (text_line[0] != '\n')
      sscanf(text_line, "%g %g %g %g %d %d",
             &(Object->Parameters->Min_U),
             &(Object->Parameters->Max_U),
             &(Object->Parameters->Min_V),
             &(Object->Parameters->Max_V),
             &(Object->Parameters->U_Steps),
             &(Object->Parameters->V_Steps));
   else
      Object->Parameters = NULL;

   /* Now read object specific information */
   switch (Object->Object_Type) {
   case Bezier_Type:
      Bezier_Wire_Reader(Object, wire_file);
      break;
   case Torus_Type:
      Torus_Wire_Reader(Object, wire_file);
      break;
   case Coil_Type:
      Coil_Wire_Reader(Object, wire_file);
      break;
   case Squishy_Type:
      Squishy_Wire_Reader(Object, wire_file);
      break;
   case Polygon_Type:
      Polygon_Wire_Reader(Object, wire_file);
      break;
   case Radial_Type:
      Radial_Wire_Reader(Object, wire_file);
      break;
   case Viewpoint_Type:
      Viewpoint_Wire_Reader(Object, wire_file);
      break;
   }
}

void
wire_write_information(Object, wire_file)
   OBJECT *Object;
   FILE *wire_file;
{
   fprintf(wire_file, "%d\n", Object->Object_Type);
   fprintf(wire_file, "%g %g %g\n",
           Object->Orientation.Origin.x,
           Object->Orientation.Origin.y,
           Object->Orientation.Origin.z);
   fprintf(wire_file, "%g %g %g\n",
           Object->Orientation.Major_Axis.x,
           Object->Orientation.Major_Axis.y,
           Object->Orientation.Major_Axis.z);
   fprintf(wire_file, "%g %g %g\n",
           Object->Orientation.Minor_Axis1.x,
           Object->Orientation.Minor_Axis1.y,
           Object->Orientation.Minor_Axis1.z);
   fprintf(wire_file, "%g %g %g\n",
           Object->Orientation.Minor_Axis2.x,
           Object->Orientation.Minor_Axis2.y,
           Object->Orientation.Minor_Axis2.z);
   fprintf(wire_file, "%g %g %g\n",
           Object->Orientation.Scale.x,
           Object->Orientation.Scale.y,
           Object->Orientation.Scale.z);
   fprintf(wire_file, "%g %g %g %g %g\n",
           Object->Texture.Color[0],
           Object->Texture.Color[1],
           Object->Texture.Color[2],
           Object->Texture.Color[3],
           Object->Texture.Phong);
   if (Object->Parameters != NULL)
      fprintf(wire_file, "%g %g %g %g %d %d\n",
              Object->Parameters->Min_U,
              Object->Parameters->Max_U,
              Object->Parameters->Min_V,
              Object->Parameters->Max_V,
              Object->Parameters->U_Steps,
              Object->Parameters->V_Steps);
   else
      fprintf(wire_file, "\n");

   /* Dump any information specific to this type of object */
   switch (Object->Object_Type) {
   case Bezier_Type:
      Bezier_Wire_Dumper(Object, wire_file);
      break;
   case Torus_Type:
      Torus_Wire_Dumper(Object, wire_file);
      break;
   case Coil_Type:
      Coil_Wire_Dumper(Object, wire_file);
      break;
   case Squishy_Type:
      Squishy_Wire_Dumper(Object, wire_file);
      break;
   case Polygon_Type:
      Polygon_Wire_Dumper(Object, wire_file);
      break;
   case Radial_Type:
      Radial_Wire_Dumper(Object, wire_file);
      break;
   case Viewpoint_Type:
      Viewpoint_Wire_Dumper(Object, wire_file);
      break;
   }
}
